@@ -58,6 +58,7 @@
 #include "config.h"
 #endif
 
+#include <xorg-server.h>
 #include <unistd.h>
 #include <misc.h>
 #include <xf86.h>
@@ -289,17 +290,6 @@ free_param_data(SynapticsPrivate *priv)
     priv->synpara = NULL;
 }
 
-static double
-synSetFloatOption(pointer options, const char *optname, double default_value)
-{
-    char *str_par;
-    double value;
-    str_par = xf86FindOptionValue(options, optname);
-    if ((!str_par) || (sscanf(str_par, "%lf", &value) != 1))
-	return default_value;
-    return value;
-}
-
 static void set_default_parameters(LocalDevicePtr local)
 {
     SynapticsPrivate *priv = local->private; /* read-only */
@@ -313,6 +303,7 @@ static void set_default_parameters(LocalDevicePtr local)
     double accelFactor;					/* 1/pixels */
     int fingerLow, fingerHigh, fingerPress;		/* pressure */
     int emulateTwoFingerMinZ;				/* pressure */
+    int emulateTwoFingerMinW;				/* width */
     int edgeMotionMinZ, edgeMotionMaxZ;			/* pressure */
     int pressureMotionMinZ, pressureMotionMaxZ;		/* pressure */
     int palmMinWidth, palmMinZ;				/* pressure */
@@ -327,7 +318,10 @@ static void set_default_parameters(LocalDevicePtr local)
     /* The synaptics specs specify typical edge widths of 4% on x, and 5.4% on
      * y (page 7) [Synaptics TouchPad Interfacing Guide, 510-000080 - A
      * Second Edition, http://www.synaptics.com/support/dev_support.cfm, 8 Sep
-     * 2008]
+     * 2008]. We use 7% for both instead for synaptics devices, and 15% for
+     * ALPS models.
+     * http://bugs.freedesktop.org/show_bug.cgi?id=21214
+     *
      * If the range was autodetected, apply these edge widths to all four
      * sides.
      */
@@ -339,8 +333,19 @@ static void set_default_parameters(LocalDevicePtr local)
         width = abs(priv->maxx - priv->minx);
         height = abs(priv->maxy - priv->miny);
         diag = sqrt(width * width + height * height);
-        ewidth = width * .04;
-        eheight = height * .055;
+        if (priv->model == MODEL_SYNAPTICS)
+        {
+            ewidth = width * .07;
+            eheight = height * .07;
+        } else if (priv->model == MODEL_ALPS)
+        {
+            ewidth = width * .15;
+            eheight = height * .15;
+        } else
+        {
+            ewidth = width * .04;
+            eheight = height * .04;
+        }
 
         l = priv->minx + ewidth;
         r = priv->maxx - ewidth;
@@ -372,15 +377,15 @@ static void set_default_parameters(LocalDevicePtr local)
 	int range = priv->maxp - priv->minp;
 
 	/* scaling based on defaults below and a pressure of 256 */
-	fingerLow = priv->minp + range * .098;
-	fingerHigh = priv->minp + range * .117;
+	fingerLow = priv->minp + range * (25.0/256);
+	fingerHigh = priv->minp + range * (30.0/256);
 	fingerPress = priv->minp + range * 1.000;
-	emulateTwoFingerMinZ = priv->minp + range * 1.1;
-	edgeMotionMinZ = priv->minp + range * .117;
-	edgeMotionMaxZ = priv->minp + range * .625;
-	pressureMotionMinZ = priv->minp + range * .117;
-	pressureMotionMaxZ = priv->minp + range * .625;
-	palmMinZ = priv->minp + range * .781;
+	emulateTwoFingerMinZ = priv->minp + range * (282.0/256);
+	edgeMotionMinZ = priv->minp + range * (30.0/256);
+	edgeMotionMaxZ = priv->minp + range * (160.0/256);
+	pressureMotionMinZ = priv->minp + range * (30.0/256);
+	pressureMotionMaxZ = priv->minp + range * (160.0/256);
+	palmMinZ = priv->minp + range * (200.0/256);
     } else {
 	fingerLow = 25;
 	fingerHigh = 30;
@@ -397,10 +402,11 @@ static void set_default_parameters(LocalDevicePtr local)
 	int range = priv->maxw - priv->minw;
 
 	/* scaling based on defaults below and a tool width of 16 */
-	palmMinWidth = priv->minw + range * .625;
-
+	palmMinWidth = priv->minw + range * (10.0/16);
+	emulateTwoFingerMinW = priv->minw + range * (7.0/16);
     } else {
 	palmMinWidth = 10;
+	emulateTwoFingerMinW = 7;
     }
 
     /* Enable tap if we don't have a phys left button */
@@ -438,6 +444,7 @@ static void set_default_parameters(LocalDevicePtr local)
     pars->fast_taps = xf86SetBoolOption(opts, "FastTaps", FALSE);
     pars->emulate_mid_button_time = xf86SetIntOption(opts, "EmulateMidButtonTime", 75);
     pars->emulate_twofinger_z = xf86SetIntOption(opts, "EmulateTwoFingerMinZ", emulateTwoFingerMinZ);
+    pars->emulate_twofinger_w = xf86SetIntOption(opts, "EmulateTwoFingerMinW", emulateTwoFingerMinW);
     pars->scroll_dist_vert = xf86SetIntOption(opts, "VertScrollDelta", horizScrollDelta);
     pars->scroll_dist_horiz = xf86SetIntOption(opts, "HorizScrollDelta", vertScrollDelta);
     pars->scroll_edge_vert = xf86SetBoolOption(opts, "VertEdgeScroll", vertEdgeScroll);
@@ -484,14 +491,14 @@ static void set_default_parameters(LocalDevicePtr local)
     pars->press_motion_min_z = xf86SetIntOption(opts, "PressureMotionMinZ", pressureMotionMinZ);
     pars->press_motion_max_z = xf86SetIntOption(opts, "PressureMotionMaxZ", pressureMotionMaxZ);
 
-    pars->min_speed = synSetFloatOption(opts, "MinSpeed", 0.4);
-    pars->max_speed = synSetFloatOption(opts, "MaxSpeed", 0.7);
-    pars->accl = synSetFloatOption(opts, "AccelFactor", accelFactor);
-    pars->trackstick_speed = synSetFloatOption(opts, "TrackstickSpeed", 40);
-    pars->scroll_dist_circ = synSetFloatOption(opts, "CircScrollDelta", 0.1);
-    pars->coasting_speed = synSetFloatOption(opts, "CoastingSpeed", 0.0);
-    pars->press_motion_min_factor = synSetFloatOption(opts, "PressureMotionMinFactor", 1.0);
-    pars->press_motion_max_factor = synSetFloatOption(opts, "PressureMotionMaxFactor", 1.0);
+    pars->min_speed = xf86SetRealOption(opts, "MinSpeed", 0.4);
+    pars->max_speed = xf86SetRealOption(opts, "MaxSpeed", 0.7);
+    pars->accl = xf86SetRealOption(opts, "AccelFactor", accelFactor);
+    pars->trackstick_speed = xf86SetRealOption(opts, "TrackstickSpeed", 40);
+    pars->scroll_dist_circ = xf86SetRealOption(opts, "CircScrollDelta", 0.1);
+    pars->coasting_speed = xf86SetRealOption(opts, "CoastingSpeed", 0.0);
+    pars->press_motion_min_factor = xf86SetRealOption(opts, "PressureMotionMinFactor", 1.0);
+    pars->press_motion_max_factor = xf86SetRealOption(opts, "PressureMotionMaxFactor", 1.0);
     pars->grab_event_device = xf86SetBoolOption(opts, "GrabEventDevice", TRUE);
 
     /* Warn about (and fix) incorrectly configured TopEdge/BottomEdge parameters */
@@ -518,9 +525,17 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     if (!priv)
 	return NULL;
 
+    /* allocate now so we don't allocate in the signal handler */
+    priv->timer = TimerSet(NULL, 0, 0, NULL, NULL);
+    if (!priv->timer) {
+	xfree(priv);
+	return NULL;
+    }
+
     /* Allocate a new InputInfoRec and add it to the head xf86InputDevs. */
     local = xf86AllocateInput(drv, 0);
     if (!local) {
+	xfree(priv->timer);
 	xfree(priv);
 	return NULL;
     }
@@ -621,6 +636,7 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     if (priv->comm.buffer)
 	XisbFree(priv->comm.buffer);
     free_param_data(priv);
+    xfree(priv->timer);
     xfree(priv);
     local->private = NULL;
     return local;
@@ -634,6 +650,9 @@ static void SynapticsUnInit(InputDriverPtr drv,
                             InputInfoPtr   local,
                             int            flags)
 {
+    SynapticsPrivate *priv = ((SynapticsPrivate *)local->private);
+    if (priv && priv->timer)
+        xfree(priv->timer);
     xfree(local->private);
     local->private = NULL;
     xf86DeleteInput(local, 0);
@@ -710,7 +729,15 @@ DeviceOn(DeviceIntPtr dev)
     xf86FlushInput(local->fd);
 
     /* reinit the pad */
-    QueryHardware(local);
+    if (!QueryHardware(local))
+    {
+        XisbFree(priv->comm.buffer);
+        priv->comm.buffer = NULL;
+        xf86CloseSerial(local->fd);
+        local->fd = -1;
+        return !Success;
+    }
+
     xf86AddEnabledDevice(local);
     dev->public.on = TRUE;
 
@@ -1942,7 +1969,8 @@ HandleState(LocalDevicePtr local, struct SynapticsHwState *hw)
     }
 
     /* Two finger emulation */
-    if (hw->z >= para->emulate_twofinger_z && hw->numFingers == 1) {
+    if (hw->numFingers == 1 && hw->z >= para->emulate_twofinger_z &&
+        hw->fingerWidth >= para->emulate_twofinger_w) {
 	hw->numFingers = 2;
     }
 
@@ -2184,6 +2212,7 @@ QueryHardware(LocalDevicePtr local)
     } else {
 	xf86Msg(X_PROBED, "%s: no supported touchpad found\n", local->name);
 	priv->proto_ops->DeviceOffHook(local);
+	return FALSE;
     }
 
     return TRUE;
